# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to morphologically analyze surface forms of Turkish words."""

from typing import Generator, List, Optional

from turkish_morphology import fst

from external.openfst import pywrapfst

_Arc = pywrapfst.Arc
_Fst = pywrapfst.Fst
_SymbolTable = pywrapfst.SymbolTable
_Weight = pywrapfst.Weight


def _input_fst(surface_form: str, symbol_table: _SymbolTable) -> _Fst:
  """Compiles given surface form in an FST.

  This function has an equivalent behaviour to the StringCompiler that
  is initialized with BYTE StringTokenType, see:

      @openfst//src/include/fst/string.h

  Args:
    surface_form: surface form to compile in an FST.
    symbol_table: symbol table which will be used to set as the symbol table
      of the compiled FST.

  Returns:
    FST that is compiled from the surface form.
  """
  input_fst = _Fst()
  last_state_index = input_fst.add_state()
  input_fst.set_start(last_state_index)

  for symbol_index in surface_form.encode("utf-8"):
    arc = _Arc(symbol_index, symbol_index, 0, last_state_index + 1)
    input_fst.add_arc(last_state_index, arc)
    last_state_index = input_fst.add_state()

  input_fst.set_final(last_state_index, 0)
  input_fst.set_input_symbols(symbol_table)
  input_fst.set_output_symbols(symbol_table)
  return input_fst


def _output_fst(analyzer_fst: _Fst, input_fst: _Fst) -> _Fst:
  """Composes input FST with the analyzer FST to create an FST with analyses."""
  input_fst.arcsort(sort_type="olabel")
  return pywrapfst.compose(input_fst, analyzer_fst)


def _extract_analyses(output_fst: _Fst,
                      state_index: int,
                      symbol_table: _SymbolTable,
                      symbols: Optional[List[str]] = []
                      ) -> Generator[str, None, None]:
  """Recursively extracts human-readable analyses from the output FST.

  This function extracts the human-readable analyses by walking over all
  possible paths from the start state of the output FST to its accept state.
  Joining the labels of the state transitions of these paths yield
  human-readable analyses.

  Args:
    output_fst: FST that is yielded after composing input and analyzer FST.
    state_index: index of an FST state from which the paths to accept state will
      be traversed.
    symbol_table: symbol table of the Turkish morphological analyzer FST.
    symbols: tokens of human-readable analyses that are gathered from the start
      state of the output FST till the state that has the given state index.

  Yields:
    Human-readable analyses that are extracted from the output FST.
  """
  arcs = list(output_fst.arcs(state_index))

  if not arcs:  # is accept state, end of human-readable analyses.
    yield "".join(symbols)

  for arc in arcs:
    new_symbols = [s for s in symbols]
    symbol = symbol_table.find(arc.olabel)

    if symbol != "<eps>":
      new_symbols.append(symbol)

    yield from _extract_analyses(
        output_fst,
        arc.nextstate,
        symbol_table,
        new_symbols,
    )


def _remove_proper_feature(human_readable: str) -> str:
  """Removes proper feature from human-readable analysis."""
  human_readable = human_readable.replace("+[Proper=False]", "")
  human_readable = human_readable.replace("+[Proper=True]", "")
  return human_readable


def surface_form(surface_form: str,
                 use_proper_feature: Optional[bool] = True) -> List[str]:
  """Morphologically analyses given surface form.

  Args:
    surface_form: surface form of a Turkish word that is to be morphologically
      analyzed.
    use_proper_feature: if true includes 'Proper' feature in the morphological
      analyses.

  Returns:
    Human-readable morphological analyses that the Turkish morphological
    analyzer yields for the given surface form. Returns an empty list if the
    given surface form is not accepted as a Turkish word form.
  """
  symbols = fst.ANALYZER.input_symbols()
  input_ = _input_fst(surface_form, symbols)
  output = _output_fst(fst.ANALYZER, input_)

  if output.start() == -1:  # has no path to the accept state.
    return []

  human_readable = _extract_analyses(output, output.start(), symbols)

  if not use_proper_feature:
    human_readable = (_remove_proper_feature(hr) for hr in human_readable)

  return sorted(set(human_readable))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: turkish_morphology/analysis.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='turkish_morphology/analysis.proto',
  package='turkish_morphology',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n!turkish_morphology/analysis.proto\x12\x12turkish_morphology\"\x18\n\x04Root\x12\x10\n\x08morpheme\x18\x01 \x01(\t\"*\n\x07\x46\x65\x61ture\x12\x10\n\x08\x63\x61tegory\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"L\n\x05\x41\x66\x66ix\x12,\n\x07\x66\x65\x61ture\x18\x01 \x01(\x0b\x32\x1b.turkish_morphology.Feature\x12\x15\n\rmeta_morpheme\x18\x02 \x01(\t\"\xb6\x01\n\x11InflectionalGroup\x12\x0b\n\x03pos\x18\x01 \x01(\t\x12&\n\x04root\x18\x02 \x01(\x0b\x32\x18.turkish_morphology.Root\x12-\n\nderivation\x18\x03 \x01(\x0b\x32\x19.turkish_morphology.Affix\x12-\n\ninflection\x18\x04 \x03(\x0b\x32\x19.turkish_morphology.Affix\x12\x0e\n\x06proper\x18\x05 \x01(\x08\"=\n\x08\x41nalysis\x12\x31\n\x02ig\x18\x01 \x03(\x0b\x32%.turkish_morphology.InflectionalGroup'
)




_ROOT = _descriptor.Descriptor(
  name='Root',
  full_name='turkish_morphology.Root',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='morpheme', full_name='turkish_morphology.Root.morpheme', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=57,
  serialized_end=81,
)


_FEATURE = _descriptor.Descriptor(
  name='Feature',
  full_name='turkish_morphology.Feature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='category', full_name='turkish_morphology.Feature.category', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='turkish_morphology.Feature.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=83,
  serialized_end=125,
)


_AFFIX = _descriptor.Descriptor(
  name='Affix',
  full_name='turkish_morphology.Affix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature', full_name='turkish_morphology.Affix.feature', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='meta_morpheme', full_name='turkish_morphology.Affix.meta_morpheme', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=127,
  serialized_end=203,
)


_INFLECTIONALGROUP = _descriptor.Descriptor(
  name='InflectionalGroup',
  full_name='turkish_morphology.InflectionalGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pos', full_name='turkish_morphology.InflectionalGroup.pos', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='root', full_name='turkish_morphology.InflectionalGroup.root', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='derivation', full_name='turkish_morphology.InflectionalGroup.derivation', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inflection', full_name='turkish_morphology.InflectionalGroup.inflection', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proper', full_name='turkish_morphology.InflectionalGroup.proper', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=206,
  serialized_end=388,
)


_ANALYSIS = _descriptor.Descriptor(
  name='Analysis',
  full_name='turkish_morphology.Analysis',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ig', full_name='turkish_morphology.Analysis.ig', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=390,
  serialized_end=451,
)

_AFFIX.fields_by_name['feature'].message_type = _FEATURE
_INFLECTIONALGROUP.fields_by_name['root'].message_type = _ROOT
_INFLECTIONALGROUP.fields_by_name['derivation'].message_type = _AFFIX
_INFLECTIONALGROUP.fields_by_name['inflection'].message_type = _AFFIX
_ANALYSIS.fields_by_name['ig'].message_type = _INFLECTIONALGROUP
DESCRIPTOR.message_types_by_name['Root'] = _ROOT
DESCRIPTOR.message_types_by_name['Feature'] = _FEATURE
DESCRIPTOR.message_types_by_name['Affix'] = _AFFIX
DESCRIPTOR.message_types_by_name['InflectionalGroup'] = _INFLECTIONALGROUP
DESCRIPTOR.message_types_by_name['Analysis'] = _ANALYSIS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Root = _reflection.GeneratedProtocolMessageType('Root', (_message.Message,), {
  'DESCRIPTOR' : _ROOT,
  '__module__' : 'turkish_morphology.analysis_pb2'
  # @@protoc_insertion_point(class_scope:turkish_morphology.Root)
  })
_sym_db.RegisterMessage(Root)

Feature = _reflection.GeneratedProtocolMessageType('Feature', (_message.Message,), {
  'DESCRIPTOR' : _FEATURE,
  '__module__' : 'turkish_morphology.analysis_pb2'
  # @@protoc_insertion_point(class_scope:turkish_morphology.Feature)
  })
_sym_db.RegisterMessage(Feature)

Affix = _reflection.GeneratedProtocolMessageType('Affix', (_message.Message,), {
  'DESCRIPTOR' : _AFFIX,
  '__module__' : 'turkish_morphology.analysis_pb2'
  # @@protoc_insertion_point(class_scope:turkish_morphology.Affix)
  })
_sym_db.RegisterMessage(Affix)

InflectionalGroup = _reflection.GeneratedProtocolMessageType('InflectionalGroup', (_message.Message,), {
  'DESCRIPTOR' : _INFLECTIONALGROUP,
  '__module__' : 'turkish_morphology.analysis_pb2'
  # @@protoc_insertion_point(class_scope:turkish_morphology.InflectionalGroup)
  })
_sym_db.RegisterMessage(InflectionalGroup)

Analysis = _reflection.GeneratedProtocolMessageType('Analysis', (_message.Message,), {
  'DESCRIPTOR' : _ANALYSIS,
  '__module__' : 'turkish_morphology.analysis_pb2'
  # @@protoc_insertion_point(class_scope:turkish_morphology.Analysis)
  })
_sym_db.RegisterMessage(Analysis)


# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to read Turkish morphological analyzer FST."""

import os
import pathlib

from external.openfst import pywrapfst

_Fst = pywrapfst.Fst

# Finite-state archive that contains the Turkish morphological analyzer FST
# is expected to be built into //src/analyzer/bin/turkish.far. See:
#     //src/analyzer/build.
_ROOT_DIR = pathlib.Path(__file__).parent.parent
_FAR_PATH = os.path.join(_ROOT_DIR, "src", "analyzer", "bin", "turkish.far")
_FST_NAME = "turkish_morphological_analyzer"


def _read_fst(far_path: str, fst_name: str) -> _Fst:
  """Reads FAR file from path and extracts the FST that has the sought name.

  Args:
    far_path: path to the finite-state archive that contains the sought FST.
    fst_name: name of the FST that will extracted from the finite-state archive.

  Returns:
    FST that has the sought name, which is extracted from the finite-state
    archive that is read from the path.
  """
  reader = pywrapfst.FarReader.open(far_path)
  return reader[fst_name]


ANALYZER = _read_fst(_FAR_PATH, _FST_NAME)



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to parse human-readable analyses into analysis protobuf messages.
"""

import re
from typing import Generator

from turkish_morphology import analysis_pb2

_Affix = analysis_pb2.Affix
_Analysis = analysis_pb2.Analysis

_AFFIX_REGEX = re.compile(
    # Derivation or inflection delimiter.
    r"[\+-]"
    # Meta-morpheme.
    r"(?P<meta_morpheme>(?:[^\W\d_]|['\.])*?)"
    # Feature category-value.
    r"\[(?P<category>[A-z]+?)=(?P<value>[A-z0-9]+?)\]")
_IG_REGEX = re.compile(
    # Beginning of an inflectional group.
    r"\("
    r"(?:"
    # Root form in first inflectional group.
    r"(?P<root>.+?)"
    # Part-of-speech tag of the first inflectional group.
    r"\[(?P<root_pos>[A-Z\.,:\(\)\'\-\"`\$]+?)\]"
    r"|"
    # Part-of-speech tag of the derived inflectional group.
    r"\[(?P<derivation_pos>[A-Z\.,:\(\)\'\-\"`\$]+?)\]"
    # Derivational morpheme and feature.
    r"(?P<derivation>-(?:[^\W\d_]|')+?\[[A-z]+?=[A-z]+?\])?"
    r")"
    # Inflectional morphemes and features.
    r"(?P<inflections>(?:\+(?:[^\W\d_]|['\.])*?\[[A-z]+?=[A-z0-9]+?\])*)"
    # End of an inflectional group.
    r"\)"
    # Optional Proper feature analysis.
    r"(?:\+\[Proper=(?P<proper>True|False)\])?")


class IllformedHumanReadableAnalysisError(Exception):
  """Raised when a human-readable analysis is structurally ill-formed."""


def _make_affix(human_readable: str) -> Generator[_Affix, None, None]:
  """Parses a sequence of human-readable affix analyses into affix protobuf.

  To illustrate, for the given human-readable analysis of below sequence of
  inflectional affixes;

      '+lAr[PersonNumber=A3pl]+Hm[Possessive=P1sg]'

  this function generates the corresponding affix protobufs;

      affix {
        feature {
          category: 'PersonNumber'
          value: 'A3pl'
        }
        meta_morpheme: 'lAr'
      }
      affix {
        feature {
          category: 'Possessive'
          value: 'P1sg'
        }
        meta_morpheme: 'Hm'
      }

  Args:
    human_readable: human-readable analysis for a sequence of derivational or
      inflectional morphemes (e.g. '-DHk[Derivation=PastNom]' or
      '+lAr[PersonNumber=A3pl]+Hm[Possessive=P1sg]+NDAn[Case=Abl]').

  Yields:
    Affix protobuf messages that are constructed from the human-readable affix
    analyses.
  """
  matches = (m.groupdict() for m in _AFFIX_REGEX.finditer(human_readable))

  for matching in matches:
    affix = _Affix()
    affix.feature.category = matching["category"]
    affix.feature.value = matching["value"]

    if matching["meta_morpheme"]:
      affix.meta_morpheme = matching["meta_morpheme"]

    yield affix


def human_readable_analysis(human_readable: str) -> _Analysis:
  """Parses given human-readable analysis into an analysis protobuf.

  To illustrate, for the given human-readable analysis;

      '(Ali[NNP]+lAr[PersonNumber=A3pl]+[Possessive=Pnon]
      +NHn[Case=Gen])+[Proper=True]'

  this function makes the corresponding analysis protobuf;

      inflectional_group {
        pos: 'NNP'
        root {
          morpheme: 'Ali'
        }
        inflection {
          feature {
            category: 'PersonNumber'
            value: 'A3pl'
          }
          meta_morpheme: 'lAr'
        }
        inflection {
          feature {
            category: 'Possessive'
            value: 'Pnon'
          }
        }
        inflection {
          feature {
            category: 'Case'
            value: 'Gen'
          }
          meta_morpheme: 'NHn'
        }
        proper: true
      }

  For the structure of the output analysis protobufs, see:

      //turkish_morphology/analysis.proto

  Args:
    human_readable: human-readable morphological analysis.

  Raises:
    IllformedHumanReadableAnalysisError: given human-readable morphological
      analysis is structurally ill-formed (e.g. missing part-of-speech tag,
      root form, derivational/inflectional morpheme, or feature category/value,
      etc.).

  Returns:
    Analysis protobuf message that is constructed from the human-readable
    analysis.
  """
  if not human_readable:
    raise IllformedHumanReadableAnalysisError(
        "Human-readable analysis is empty.")

  igs = tuple(_IG_REGEX.finditer(human_readable))
  matches = [ig.groupdict() for ig in igs]

  if not (igs and len(human_readable) == igs[-1].end() and matches[0]["root"]
          and matches[0]["root_pos"]
          and all(m["derivation"] for m in matches[1:])
          and all(m["derivation_pos"] for m in matches[1:])):
    raise IllformedHumanReadableAnalysisError(
        f"Human-readable analysis is ill-formed: '{human_readable}'")

  analysis = _Analysis()

  for position, matching in enumerate(matches):
    ig = analysis.ig.add()

    if position == 0:
      ig.pos = matching["root_pos"]
      ig.root.morpheme = matching["root"]
    else:
      ig.pos = matching["derivation_pos"]
      derivation = tuple(_make_affix(matching["derivation"]))[0]
      ig.derivation.CopyFrom(derivation)

    inflections = _make_affix(matching["inflections"])
    ig.inflection.extend(inflections)

    if matching["proper"]:
      ig.proper = matching["proper"] == "True"

  return analysis
